/*  Copyright (C) 2010  Faculdade de Ciências e Tecnologia
 *                      Universidade Nova de Lisboa
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package pt.unl.fct.di.jtraceview.trace;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;

import pt.unl.fct.di.jtraceview.trace.events.TxAbortEvent;
import pt.unl.fct.di.jtraceview.trace.events.TxCommitEvent;
import pt.unl.fct.di.jtraceview.trace.events.TxReadEvent;
import pt.unl.fct.di.jtraceview.trace.events.TxStartEvent;
import pt.unl.fct.di.jtraceview.trace.events.TxWriteEvent;
import pt.unl.fct.di.jtraceview.trace.events.TxAbortEvent.AbortType;

public class TraceFileReader {
	
	private RandomAccessFile raf;
	
	private static final int EVENT_SIZE = 23;	
	private int eventBUFSIZE = 2*1024*1024;
	private int byteBUFSIZE  = EVENT_SIZE*eventBUFSIZE; // 46 MB
	
	private byte[]  byteBuff; 	// read byte buffer
	private Event[] eventBuff;	// buffed Events
	
	private int firstEvent = -1;	// number first event in the buff
	private int currEvent  = -1;	// number of the current event
	private int endEvent   = -1;	// number last event in the buff
	private int maxEvent;			// number the lastEvent
	// firstEvent equals the firstEvent*EVENT_SIZE pos in the file
	// same for currEvent, endEvent

	public TraceFileReader(File traceFile) {
		try {
			raf = new RandomAccessFile(traceFile, "r");

			if (byteBUFSIZE > traceFile.length()) {
				byteBUFSIZE = (int)traceFile.length();
				eventBUFSIZE = byteBUFSIZE  / EVENT_SIZE;
			}

			byteBuff  = new byte[byteBUFSIZE];
			eventBuff = new Event[eventBUFSIZE];
			reset();
			maxEvent = (int)(traceFile.length() / EVENT_SIZE);
			
			System.out.print("Initial buffer: ");

			if (fillEventBuffer(0))
				System.out.println("done.");
			else
				System.out.println("fail.");
			
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		}
	}
	
	// numEv = the number of the Event we want
	public Event getEvent(int numEv) {
		boolean r = true;
		
		if (numEv<firstEvent || numEv>=endEvent){
			if (numEv<0 || numEv>=maxEvent)
				return null;
			else
				r = fillEventBuffer(numEv);
		}
	
		currEvent = numEv;
		if (r) return eventBuff[numEv-firstEvent];
		else return null;
	}
		
	
	// start - the number of the Event from where to start filling the buffer
	// returns the success of the operation
	public boolean fillEventBuffer (int start) {
		if (raf == null)
			return false;
		
		int nRead;
//		if (start<firstEvent)
//			System.out.println("start: "+start+", firstEvent: "+firstEvent);
		start -= 100000;	// this way we don't need to keep losing the buffer when we read some events backwards
		if (start<0) start = 0;
		
		long newpos = start*EVENT_SIZE;
		
		try {
			// invalid position in the file
			if ((newpos < 0)||((newpos+EVENT_SIZE) > raf.length()))
				return false;
			
			raf.seek(newpos);

			nRead = raf.read(byteBuff);
			if (nRead%EVENT_SIZE!=0)
				System.out.println("Warning in fillEventBuffer: nRead is not %EVENT_SIZE==0");
			nRead = nRead/EVENT_SIZE; // number of events read
			
			firstEvent = start;
			endEvent   = firstEvent + nRead;
			currEvent  = firstEvent;
			
			int index = 0;
			for (int i=0; i<nRead; i++, index+=EVENT_SIZE)
				eventBuff[i] = genEvent( byteBuff , index );

		} catch (IOException e) {
			e.printStackTrace();
			return false;
		}
		
		
		return true;
	}


	public Event getPreviousEvent() {
		return getEvent( currEvent-1 );
	}
	
	public Event getCurrentEvent() {
		return getEvent( currEvent );
	}
	
	public Event getNextEvent() {
		return getEvent( currEvent+1 );
	}
	
	public Event getLastEvent() {
		
		try {
			long pos = raf.getFilePointer();
			raf.seek(raf.length()-EVENT_SIZE);
			byte[] byteb = new byte[EVENT_SIZE];
			raf.read(byteb);
			raf.seek(pos);
			return genEvent(byteb,0);
		} catch (IOException e) {
			e.printStackTrace();
			return null;
		}
		
//		return getEvent( maxEvent-1 );
	}
	
	public void reset() {
		currEvent = -1;
	}
	
	public long setSavePoint() {
		return currEvent;
	}
	
	public void restoreSavePoint(long savePoint) {
		currEvent = (int)savePoint;
	}
	
	public void close() {
		try {
			raf.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	public static Event genEvent(byte[] ev, int pos) {
		Event e = null;
		if (pos<0) return null;
		
		EventId evId = getEventId(ev,pos);
		
		short threadId = getThreadId(ev,pos);
		int txId       = getTxId(ev,pos);
		long timestamp = getTimestamp(ev,pos);
		
		if (evId == EventId.TX_START) {
			e = new TxStartEvent(timestamp, threadId, txId);
		}
		else if (evId == EventId.TX_COMMIT) {
			e = new TxCommitEvent(timestamp, threadId, txId);
		}
		else if (evId == EventId.TX_ABORT) {
			e = new TxAbortEvent(timestamp, threadId, txId, getAbortType(ev,pos));
		}
		else if (evId == EventId.TX_READ) {
			e = new TxReadEvent(timestamp, threadId, txId, getVar(ev,pos)+"");
		}
		else if (evId == EventId.TX_WRITE) {
			e = new TxWriteEvent(timestamp, threadId, txId, getVar(ev,pos)+"");
		}
		
		e.threadId = threadId;
		
		return e;
	}
	
	private static int getTxId(byte[] ev, int pos) {
		int res = 0;
		
		res |= ((int)ev[11+pos] & 0xFF) << 0;
		res |= ((int)ev[12+pos] & 0xFF) << 8;
		res |= ((int)ev[13+pos] & 0xFF) << 16;
		res |= ((int)ev[14+pos] & 0xFF) << 24;
		
		return res;
	}
	
	private static AbortType getAbortType(byte[] ev, int pos) {
		return AbortType.getType(ev[15+pos]);
	}
	
	private static long getVar(byte[] ev, int pos) {
		long res = 0L;
		
		res |= ((long)ev[15+pos] & 0xFFL) << 0;
		res |= ((long)ev[16+pos] & 0xFFL) << 8;
		res |= ((long)ev[17+pos] & 0xFFL) << 16;
		res |= ((long)ev[18+pos] & 0xFFL) << 24;
		res |= ((long)ev[19+pos] & 0xFFL) << 32;
		res |= ((long)ev[20+pos] & 0xFFL) << 40;
		res |= ((long)ev[21+pos] & 0xFFL) << 48;
		res |= ((long)ev[22+pos] & 0xFFL) << 56;
		
		return res;
	}
	private static EventId getEventId(byte[] ev, int pos) {
		return EventId.getId(ev[8+pos]);
	}
	
	private static short getThreadId(byte[] ev, int pos) {
		short res = 0;
		
		res |= ((short)ev[9+pos] & 0xFF) << 0;
		res |= ((short)ev[10+pos] & 0xFF) << 8;
		
		return res;
	}
	
	private static long getTimestamp(byte[] ev, int pos) {
		long res = 0L;
		
		res |= ((long)ev[0+pos] & 0xFFL) << 0;
		res |= ((long)ev[1+pos] & 0xFFL) << 8;
		res |= ((long)ev[2+pos] & 0xFFL) << 16;
		res |= ((long)ev[3+pos] & 0xFFL) << 24;
		res |= ((long)ev[4+pos] & 0xFFL) << 32;
		res |= ((long)ev[5+pos] & 0xFFL) << 40;
		res |= ((long)ev[6+pos] & 0xFFL) << 48;
		res |= ((long)ev[7+pos] & 0xFFL) << 56;
		
		return res;
	}
	
	public static void main(String[] args) throws Exception {
		TraceFileReader tfr = new TraceFileReader(new java.io.File(args[0]));
		Event e = tfr.getCurrentEvent();
		System.out.println(e);
		while((e = tfr.getNextEvent()) != null) {
			System.out.println(e);
		}
	}

}
